/*
This  program  and documentation is Copyrighted by its author, Carey
Bloodworth, on July 18, 1998.   You  may  do what ever you wish with
them, with two exceptions.   First,  you  have  to leave my name and
this notice on it, and second, you may  not  put  the  infamous  GNU
'copyleft'  on  it.   The  GNU 'copyleft' is far too restrictive and
quite 'militant' in its wording.  I don't approve of it and I do not
want my  code  being  inflicted  by  it.   Beyond  that,  you can do
whatever you want with it.  You may use it in any  application,  for
any reason, without informing me.

Carey Bloodworth

*/
/*
**
**
** The main reason it's not in the Public Domain is  because  I  put
** quite a bit of work  into  these  and  I'm rather  proud  of  the
** results, and I just want my name left in. (I guess you could call
** it  pride, or  perhaps  ego.)   Beyond  that, you can do whatever
** you want with them.   You  can  use  them  for  any purpose.  You
** don't have to ask me for permission or notify  me  that  you  are
** using it.
**
** I'd also appreciate it if you see any ideas in my program and
** implement them in yours, that you put a note in it about it.
** You don't have to, though.  It's just nice to see your own name
** in somebody's list of 'thank you's.
*/
/*
** The Fast Fourier Transforms.  I call this type of optimized
** Radix-2 FFT a "Quad-2" FFT, due to it doing all four (ie: quad)
** quadrants at the same time.
**
** Any decent code optimizer will deal with us passing -1,0,+1 as
** the Pow_r (Pr) sign (Sr) and Pow_i (Pi) sign (Si).  Plus, this
** lets the code optimizer know what sign the powers will be and
** generate more appropriate code.
*/

extern /*long*/ double SineTable[];

int Log2(int Num) {
	int x=-1;
	if (Num==0) return 0;
	while (Num) {
		x++;
		Num/=2;
	}
return x;
}

#define NextTrigPow(Pr,Pi,Nr,Ni)    \
 {/*long*/ double temp;             \
   temp = Pr;                       \
   Pr = Pr * Nr - Pi   * Ni + Pr;   \
   Pi = Pi * Nr + temp * Ni + Pi;   \
 }


/***********************************
** Decimation in time transforms. **
************************************/

/* The iterative butterfly loop. */
#define ButterflyLoop(Ndx,Sr,Pr,Si,Pi)                   \
   {Cmplx *Left,*Right;                                  \
    Left=&Data[(Ndx)];                                   \
    Right=Left+Step2;                                    \
    while (Left < LastPoint)                             \
      {                                                  \
       {double tr, ti, t;                                \
        tr = (ti = Right->r) * (Pr) * (Sr) -             \
             Right->i * (Pi) * (Si);                     \
        Right->r = (t = Left->r) - tr;                   \
        Left->r  = t + tr;                               \
        ti = ti * (Pi) * (Si) +                          \
             Right->i * (Pr) *(Sr);                      \
        Right->i = (t = Left->i) - ti;                   \
        Left->i  = t + ti;                               \
       }                                                 \
       Left += Step; Right += Step;                      \
      }                                                  \
   }

/* The butterfly for the recursive FFTs */
#define RButterfly(Ndx,Sr,Pr,Si,Pi)                   \
   {int z=(Ndx);                                      \
    {double tr, ti, t;                                \
     tr = (ti = Right[z].r) * (Pr) * (Sr) -           \
          Right[z].i * (Pi) * (Si);                   \
     Right[z].r = (t = Left[z].r) - tr;               \
     Left[z].r  = t + tr;                             \
     ti = ti * (Pi) * (Si) +                          \
          Right[z].i * (Pr) *(Sr);                    \
     Right[z].i = (t = Left[z].i) - ti;               \
     Left[z].i  = t + ti;                             \
    }                                                 \
   }

static void RevFFT_T(Cmplx *Data, int Len) {
/* Reverse Decimation in Time */

	int Step, Step2, Step4, Step8;
	/* long */double Nth_r,Nth_i;
	/* long */double Pow_r,Pow_i;
	int b;
	int TrigIndex=1;
	Cmplx *LastPoint=&Data[Len];

	Step = 1;
	while (Step < Len) {
		Step *= 2;
		Step2 = Step/2;
		Step4 = Step/4;
		Step8 = Step/8;

		ButterflyLoop(0,POS,1.0,NEG,0.0);

		Pow_r = 1.0;Pow_i = 0.0;
		Nth_i = SineTable[TrigIndex++];
		Nth_r = -2.0*SineTable[TrigIndex]*SineTable[TrigIndex];
		for (b = 1; b < Step8; b++) {
			/*Cmplx *Left,*Right;*/
			NextTrigPow(Pow_r,Pow_i,Nth_r,Nth_i);
			ButterflyLoop(b,      POS,Pow_r,NEG,Pow_i);
			ButterflyLoop(Step2-b,NEG,Pow_r,NEG,Pow_i);
			ButterflyLoop(Step4-b,POS,Pow_i,NEG,Pow_r);
			ButterflyLoop(Step4+b,NEG,Pow_i,NEG,Pow_r);
		}
		if (Step8) {
			/*long*/ double sq=M_SQRT_2;
			ButterflyLoop(Step8,      POS,sq,NEG,sq);
			ButterflyLoop(Step2-Step8,NEG,sq,NEG,sq);
        }
		if (Step4) ButterflyLoop(Step4,POS,0.0,NEG,1.0);
    }
}

static void RevRFFT_T(Cmplx *Data,int Len) {
	/*
	** Recursive Reverse Decimation in Time
	*/
	int x,Len2,Len4,Len8;
	Cmplx *Left,*Right;
	/* long */ double Pow_r=1.0,Pow_i=0.0;
	/* long */ double Nth_r,Nth_i;

	Len2 = Len/2;Len4 = Len/4;Len8 = Len/8;
	if (Len<=(CACHE_SIZE/sizeof(Cmplx))) {
		RevFFT_T(Data,Len);
		return;
	}
	if (Len2 >= 2) {
		RevRFFT_T(Data,Len2);
		RevRFFT_T(Data+Len2,Len2);
	}

	x=Log2(Len);
	Nth_r=-2.0*SineTable[x+1]*SineTable[x+1];
	Nth_i=SineTable[x];

	Left=&Data[0];
	Right=&Data[Len2];
	RButterfly(0,POS,1.0,NEG,0.0);
	for (x=1;x<Len8;x++) {
		NextTrigPow(Pow_r,Pow_i,Nth_r,Nth_i);
		RButterfly(x,     POS,Pow_r,NEG,Pow_i);
		RButterfly(Len2-x,NEG,Pow_r,NEG,Pow_i);
		RButterfly(Len4-x,POS,Pow_i,NEG,Pow_r);
		RButterfly(Len4+x,NEG,Pow_i,NEG,Pow_r);
	}

	if (Len8) {
		/*long*/double sq=M_SQRT_2;
		RButterfly(Len8,     POS,sq,NEG,sq);
		RButterfly(Len2-Len8,NEG,sq,NEG,sq);
	}
	if (Len4)
		RButterfly(Len4,POS,0,NEG,1.0);
}

static void FwdFFT_T(Cmplx *Data, int Len) {
	/* Forward Decimation in Time */
	int Step, Step2, Step4, Step8;
	/* long */double Nth_r,Nth_i;
	/* long */double Pow_r,Pow_i;
	int b;
	int TrigIndex=1;
	Cmplx *LastPoint=&Data[Len];

	Step = 1;
	/* save a wee bit of time since the data is zero padded */
	#ifdef ZERO_PAD_CUT
		Step = 2;
		TrigIndex=2;
		#ifdef CUT3
			Step = 4;
			TrigIndex=3;
		#endif
		#ifdef CUT4
			Step = 8;
			TrigIndex=4;
		#endif
	#endif
	while (Step < Len) {
		Step *= 2;
		Step2 = Step/2;
		Step4 = Step/4;
		Step8 = Step/8;

		ButterflyLoop(0,POS,1.0,POS,0.0);

		Pow_r = 1.0;
		Pow_i = 0.0;
		Nth_i = SineTable[TrigIndex++];
		Nth_r = -2.0*SineTable[TrigIndex]*SineTable[TrigIndex];
		for (b = 1; b < Step8; b++) {
			/*Cmplx *Left,*Right;*/
			NextTrigPow(Pow_r,Pow_i,Nth_r,Nth_i);
			ButterflyLoop(b,      POS,Pow_r,POS,Pow_i);
			ButterflyLoop(Step2-b,NEG,Pow_r,POS,Pow_i);
			ButterflyLoop(Step4-b,POS,Pow_i,POS,Pow_r);
			ButterflyLoop(Step4+b,NEG,Pow_i,POS,Pow_r);
		}
		if (Step8) {
			/*long*/ double sq=M_SQRT_2;
			ButterflyLoop(Step8,      POS,sq,POS,sq);
			ButterflyLoop(Step2-Step8,NEG,sq,POS,sq);
		}
		if (Step4)
			ButterflyLoop(Step4,POS,0.0,POS,1.0);
	}
}

static void FwdRFFT_T(Cmplx *Data,int Len) {
	/*
	** Recursive Forward Decimation in Time
	*/
	int x,Len2,Len4,Len8;
	Cmplx *Left,*Right;
	/* long */ double Pow_r=1.0,Pow_i=0.0;
	/* long */ double Nth_r,Nth_i;

	Len2 = Len/2;Len4 = Len/4;Len8 = Len/8;
	if (Len<=(CACHE_SIZE/sizeof(Cmplx))) {
		FwdFFT_T(Data,Len);
		return;
	}
	else
		if (Len2 >= 2) {
			FwdRFFT_T(Data,Len2);
			FwdRFFT_T(Data+Len2,Len2);
		}

	x=Log2(Len);
	Nth_r=-2.0*SineTable[x+1]*SineTable[x+1];
	Nth_i=SineTable[x];

	Left=&Data[0];
	Right=&Data[Len2];
	RButterfly(0,POS,1.0,POS,0.0);
	for (x=1;x<Len8;x++) {
		NextTrigPow(Pow_r,Pow_i,Nth_r,Nth_i);
		RButterfly(x,     POS,Pow_r,POS,Pow_i);
		RButterfly(Len2-x,NEG,Pow_r,POS,Pow_i);
		RButterfly(Len4-x,POS,Pow_i,POS,Pow_r);
		RButterfly(Len4+x,NEG,Pow_i,POS,Pow_r);
	}

	if (Len8) {
		/*long*/double sq=M_SQRT_2;
		RButterfly(Len8,     POS,sq,POS,sq);
		RButterfly(Len2-Len8,NEG,sq,POS,sq);
	}
	if (Len4)
		RButterfly(Len4,POS,0,POS,1.0);
}

/*
** Reorder complex data array by bit reversal rule.
** It's faster if we treat it as two doubles, instead
** of a struct Cmplx.
*/
void FFTReOrder(double *point, int Len) {
	int Index, xednI, x;
	double temp;

	xednI = 0;
	for (Index = 0; Index < Len; Index += 2) {
		if (Index < xednI) {
			temp = point[Index];
			point[Index] = point[xednI];
			point[xednI] = temp;
			temp = point[Index + 1];
			point[Index + 1] = point[xednI + 1];
			point[xednI + 1] = temp;
		}
		x = Len / 2;
		while ((x > 0) && (x <= xednI)) {
			xednI -= x;
			x /= 2;
		}
		xednI += x;
	}
}

void RealFFT(double *ddata,int Len, int Dir) {
/*
** I don't have a Real Value FFT, so I have to fake it
** by putting a wrapper around a complex FFT.
*/
	double Pow_r, Pow_i, Nth_r, Nth_i, temp;
	int i, j;
	Cmplx *Data=(Cmplx*)ddata;
	int Half;
	double NegHalf, PosHalf;

	Len /= 2;       /* Len/2 Cmplx data */
	Half = Len/2;
	NegHalf = -0.5;
	PosHalf = 0.5;
	if (Dir > 0) {
		NegHalf = 0.5;
		PosHalf = -0.5;
		#if !defined(ZERO_PAD_CUT) || (defined(ZERO_PAD_CUT) && defined(CUT1))
			FFTReOrder(ddata,Len*2);
		#endif
		FwdRFFT_T(Data,Len);
	}

	i=Log2(Len)+1;
	Nth_r = -2.0*SineTable[i+1]*SineTable[i+1];
	Nth_i = Dir*SineTable[i];
	Pow_r = 1 + Nth_r;
	Pow_i = Nth_i;
	for (i = 1, j = Len - i; i < Half; i++, j--) {
		double p1r,p1i,p2r,p2i;
		/* Seperate the two points from the jumbled points */
		p1r =     0.5 * (Data[i].r + Data[j].r);
		p2i = PosHalf * (Data[i].r - Data[j].r);
		p2r = NegHalf * (Data[i].i + Data[j].i);
		p1i =     0.5 * (Data[i].i - Data[j].i);
		/* Almost a standard Decimation in Time butterfly... */
		{
			double tmp;
			tmp = p2r;
			p2r = p2r * Pow_r - p2i * Pow_i;
			p2i = p2i * Pow_r + tmp * Pow_i;
		}
		Data[i].r = p1r + p2r;
		Data[i].i = p1i + p2i;
		Data[j].r = p1r - p2r;
		Data[j].i = -(p1i - p2i); /* ... except this is negated */
		NextTrigPow(Pow_r,Pow_i,Nth_r,Nth_i);
	}

	temp = Data[0].r;
	Data[0].r = temp + Data[0].i;
	Data[0].i = temp - Data[0].i;

	if (Dir < 0) {
		Data[0].r /= 2.0;
		Data[0].i /= 2.0;
		FFTReOrder(ddata,Len*2);
		RevRFFT_T(Data,Len);
	}
}


void FwdRealFFT(double *ddata,int Len) {
	RealFFT(ddata,Len,1);
}

void RevRealFFT(double *ddata,int Len) {
	RealFFT(ddata,Len,-1);
}



void InitFFT(unsigned long int Len) {
/*
** If you need to use your own trig tables, you can calculate
** them here.
*/
	int x;
	unsigned long int P;
	#if (LDBL_DIG > 18) || (LDBL_MANT_DIG > 64)
		SineTable[0]=0.0;
		x=1;
		P=1;
		while (P<=Len*4) {
			SineTable[x]=sin(M_PI/P);
			P*=2;
			x++;
		}
	#endif
}


/*
** Precalculated trig in long double (64 bits of mantissa).
** Perhaps not accurate to the last bit, but it's good enough.
*/
/*long*/ double SineTable[]={
/* sin(M_PI/**  ) */      0.000000000000000000000000000000000000000e+00L,
/* sin(M_PI/1   ) */      0.000000000000000000000000000000000000000e+00L,
/* sin(M_PI/2   ) */      1.000000000000000000000000000000000000000e+00L,
/* sin(M_PI/4   ) */      7.071067811865475027412186737052479656995e-01L,
/* sin(M_PI/8   ) */      3.826834323650897575798401906155277174548e-01L,
/* sin(M_PI/16  ) */      1.950903220161282603309360617060974618653e-01L,
/* sin(M_PI/32  ) */      9.801714032956059818174621156572356994729e-02L,
/* sin(M_PI/64  ) */      4.906767432741801234115375240918410781887e-02L,
/* sin(M_PI/128 ) */      2.454122852291228707409531661909340982675e-02L,
/* sin(M_PI/256 ) */      1.227153828571992560101701352781589093865e-02L,
/* sin(M_PI/512 ) */      6.135884649154475120776119911880641666357e-03L,
/* sin(M_PI/1k  ) */      3.067956762965976150181468540267815114930e-03L,
/* sin(M_PI/2k  ) */      1.533980186284765552615777517431183696317e-03L,
/* sin(M_PI/4k  ) */      7.669903187427044969921158257264437452250e-04L,
/* sin(M_PI/8k  ) */      3.834951875713955740925670268026692610874e-04L,
/* sin(M_PI/16k ) */      1.917475973107032999678822626776764082024e-04L,
/* sin(M_PI/32k ) */      9.587379909597734213219655252657958044438e-05L,
/* sin(M_PI/64k ) */      4.793689960306688268003999509048185245774e-05L,
/* sin(M_PI/128k) */      2.396844980841821779338901565736819065933e-05L,
/* sin(M_PI/256k) */      1.198422490506970595472088780830688392598e-05L,
/* sin(M_PI/512k) */      5.992112452642427609071640315363538320526e-06L,
/* sin(M_PI/1m  ) */      2.996056226334660633689455089267994480906e-06L,
/* sin(M_PI/2m  ) */      1.498028113169011170304617541759739651752e-06L,
/* sin(M_PI/4m  ) */      7.490140565847156918673210856951527603087e-07L,
/* sin(M_PI/8m  ) */      3.745070282923841092784580930619142691285e-07L,
/* sin(M_PI/16m ) */      1.872535141461953375578708413939921229030e-07L,
/* sin(M_PI/32m ) */      9.362675707309807915379451515036635100842e-08L,
/* sin(M_PI/64m ) */      4.681337853654909086833363351942693952878e-08L,
/* sin(M_PI/128m) */      2.340668926827455184830686918395770135248e-08L
};

void ReOrderShort(Short *Num, int Len) {
	int Index,xednI,k;
	int Len2=Len/2;

	xednI = 0;
	for (Index = 0;Index < Len;Index+=2) {
		if (xednI > Index) {
			Short t;
			t=Num[xednI];
			Num[xednI] = Num[Index];
			Num[Index] = t;
			t=Num[xednI+1];
			Num[xednI+1] = Num[Index+1];
			Num[Index+1] = t;
		}
		k=Len2;
		while ((k <= xednI) && (k >=1)) {
			xednI-=k;
			k/=2;
		}
		xednI+=k;
	}
}

/*
** For loading our numbers into the FFTNum array.
** Just part of a 'zero-cut' optimization.  Since the data
** is short int's, we can take advantage of the cheaper scrambling,
** and the easier manipulation and go ahead and do up to three
** passes of the FFT!
*/
void LoadFFTNum(Short *Num, double *FFTNum, int Len) {
	int x,Len2=Len*2;
	int y;
	#ifdef ZERO_PAD_CUT
		#ifdef CUT1
			for (x = 0; x < Len; x++)
				FFTNum[x] = FFTNum[x+Len] = Num[Len - x - 1];
		#endif
		#ifdef CUT2
			ReOrderShort(Num,Len);
			y=Len-1;
			x=0;
			while (x < Len2) {
				int r,i;
				r=Num[y];
				i=Num[y-1];
				y-=2;
				FFTNum[x+0]=r;
				FFTNum[x+1]=i;
				FFTNum[x+2]=r;
				FFTNum[x+3]=i;
				x+=4;
			}
			ReOrderShort(Num,Len);
		#endif
		#ifdef CUT3
			ReOrderShort(Num,Len);
			y=Len-1;
			x=0;
			while (x < Len2) {
				int r1,i1,r2,i2;
				r1=Num[y];i1=Num[y-1];y-=2;
				r2=Num[y];i2=Num[y-1];y-=2;
				FFTNum[x+0]=r1+r2;
				FFTNum[x+1]=i1+i2;
				FFTNum[x+2]=r1+ -i2;
				FFTNum[x+3]=i1+r2;
				FFTNum[x+4]=r1-r2;
				FFTNum[x+5]=i1-i2;
				FFTNum[x+6]=r1- -i2;
				FFTNum[x+7]=i1-r2;
				x+=8;
			}
			ReOrderShort(Num,Len);
		#endif
		#ifdef CUT4
			if (Len <8) {printf("Too small.\n");exit(1);}
				ReOrderShort(Num,Len);
				y=Len-1;
				x=0;
				while (y > 0) {
					#if INT_MAX==32767
     					long    /* Just for testing under Borland C */
					#else
						/* short */
					#endif
					int r1,i1,r2,i2,r3,i3,r4,i4;
					r1=Num[y];i1=Num[y-1];y-=2;
					r2=Num[y];i2=Num[y-1];y-=2;
					r3=Num[y];i3=Num[y-1];y-=2;
					r4=Num[y];i4=Num[y-1];y-=2;
					FFTNum[ 0]=(r1+r2) + (r3+r4);
					FFTNum[ 8]=(r1+r2) - (r3+r4);
					FFTNum[ 4]=(r1-r2) - (i3-i4);
					FFTNum[12]=(r1-r2) + (i3-i4);
					FFTNum[ 5]=(i1-i2) + (r3-r4);
					FFTNum[13]=(i1-i2) - (r3-r4);
					FFTNum[ 1]=(i1+i2) + (i3+i4);
					FFTNum[ 9]=(i1+i2) - (i3+i4);
					FFTNum[ 2]=(r1-i2) + ((r3-i4)-(i3+r4))*M_SQRT_2;
					FFTNum[10]=(r1-i2) - ((r3-i4)-(i3+r4))*M_SQRT_2;
					FFTNum[ 6]=(r1+i2) - ((r3+i4)+(i3-r4))*M_SQRT_2;
					FFTNum[14]=(r1+i2) + ((r3+i4)+(i3-r4))*M_SQRT_2;
					FFTNum[ 7]=(i1-r2) + ((r3+i4)-(i3-r4))*M_SQRT_2;
					FFTNum[15]=(i1-r2) - ((r3+i4)-(i3-r4))*M_SQRT_2;

					FFTNum[ 3]=(i1+r2) + ((i3+r4)+(r3-i4))*M_SQRT_2;
					FFTNum[11]=(i1+r2) - ((i3+r4)+(r3-i4))*M_SQRT_2;
					FFTNum+=16;
				}
				ReOrderShort(Num,Len);
		#endif
	#else
		for (x = 0; x < Len; x++)    FFTNum[x] = Num[Len - x - 1];
		for (x = Len; x < Len2; x++) FFTNum[x] = 0.0;
	#endif
}
